/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <bsd/string.h>

#include <rte_sft.h>

#include <doca_argp.h>
#include <doca_dpi.h>
#include <doca_log.h>

#include <utils.h>
#include <sig_db.h>

#include "ips_core.h"

DOCA_LOG_REGISTER(IPS::Core);

struct doca_dpi_ctx *dpi_ctx;	/* Shared DPI context across all workers */

void
signal_handler(int signum)
{
	if (signum == SIGINT || signum == SIGTERM) {
		DOCA_LOG_INFO("Signal %d received, preparing to exit...", signum);
		force_quit = true;
	}
}

/*
 * Callback function for DPI Worker thread, called when a packet is matched.
 * The action to take is determined by the DPI rule, DROP OR RSS.
 *
 * @queue [in]: Queue that handled the packet
 * @dpi_result [in]: Result of the DPI match
 * @fid [in]: Flow ID of the packet
 * @user_data [in]: User data passed to the callback function
 * @action [out]: Action to take after the callback function returns
 * @return: 0 on success and negative value on failure
 */
static int
dpi_worker_on_match(int queue, const struct doca_dpi_result *dpi_result, uint32_t fid,
			void *user_data, enum dpi_worker_action *action)
{
	uint32_t sig_id = dpi_result->info.sig_id;
	struct ips_config *ips = (struct ips_config *) user_data;
	struct doca_dpi_sig_data sig_data;
	struct sig_info *data;
	bool blocked = false;
	doca_error_t result;
	int ret;

	ret = doca_dpi_signature_get(dpi_ctx, dpi_result->info.sig_id, &sig_data);
	if (ret != 0)
		return -1;

	if (sig_db_sig_info_get(sig_id, &data) != DOCA_SUCCESS)
		result = sig_db_sig_info_create(sig_id, sig_data.name, dpi_result->info.action == DOCA_DPI_SIG_ACTION_DROP);
	else
		result = sig_db_sig_info_set(sig_id, sig_data.name);

	if (result != DOCA_SUCCESS)
		return -1;

	result = sig_db_sig_info_fids_inc(sig_id);
	if (result != DOCA_SUCCESS)
		return -1;

	result = sig_db_sig_info_get_block_status(sig_id, &blocked);
	if (result != DOCA_SUCCESS)
		return -1;

	if (ips->print_on_match)
		printf_signature(dpi_ctx, sig_id, fid, blocked);

	*action = blocked ? DPI_WORKER_DROP : DPI_WORKER_RSS_FLOW;
	return 0;
}

/*
 * ARGP Callback - Handle CDO file path parameter
 *
 * @param [in]: Input parameter
 * @config [in/out]: Program configuration context
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
cdo_file_callback(void *param, void *config)
{
	struct ips_config *ips = (struct ips_config *) config;
	char *cdo_file_path = (char *) param;

	if (strnlen(cdo_file_path, MAX_FILE_NAME) == MAX_FILE_NAME) {
		DOCA_LOG_ERR("CDO file name is too long - MAX=%d", MAX_FILE_NAME - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (access(cdo_file_path, F_OK) == -1) {
		DOCA_LOG_ERR("CDO file not found %s", cdo_file_path);
		return DOCA_ERROR_NOT_FOUND;
	}
	strlcpy(ips->cdo_filename, cdo_file_path, MAX_FILE_NAME);
	return DOCA_SUCCESS;
}

/*
 * ARGP Callback - Handle output CSV path parameter
 *
 * @param [in]: Input parameter
 * @config [in/out]: Program configuration context
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
csv_file_callback(void *param, void *config)
{
	struct ips_config *ips = (struct ips_config *) config;
	char *value = (char *) param;
	uint32_t len = strnlen(value, MAX_FILE_NAME);

	if (len == 0)
		return DOCA_SUCCESS;
	if (len >= MAX_FILE_NAME) {
		DOCA_LOG_ERR("CSV file name is too long - MAX=%d", MAX_FILE_NAME - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}
	strlcpy(ips->csv_filename, value, MAX_FILE_NAME);
	ips->create_csv = true;
	DOCA_LOG_DBG("Enabling CSV report [%s]", ips->csv_filename);
	return DOCA_SUCCESS;
}

/*
 * ARGP Callback - Handle Netflow source ID parameter
 *
 * @param [in]: Input parameter
 * @config [in/out]: Program configuration context
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
netflow_callback(void *param, void *config)
{
	struct ips_config *ips = (struct ips_config *) config;

	ips->netflow_source_id =  *(int *) param;
	return DOCA_SUCCESS;
}

/*
 * ARGP Callback - Handle print on match parameter
 *
 * @param [in]: Input parameter
 * @config [in/out]: Program configuration context
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
print_match_callback(void *param, void *config)
{
	struct ips_config *ips = (struct ips_config *) config;

	DOCA_LOG_DBG("Enabling print on match");
	ips->print_on_match = *(bool *) param;
	return DOCA_SUCCESS;
}

/*
 * ARGP Callback - Handle fragmented packet parameter
 *
 * @param [in]: Input parameter
 * @config [in/out]: Program configuration context
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
frag_callback(void *param, void *config)
{
	struct ips_config *ips = (struct ips_config *) config;

	ips->dpdk_config->sft_config.enable_frag = *(bool *) param;
	return DOCA_SUCCESS;
}

int
ips_init(const struct application_dpdk_config *dpdk_config,
	struct ips_config *ips_config, struct dpi_worker_attr *dpi_worker)
{
	doca_error_t result;
	int ret;
	struct doca_dpi_config_t doca_dpi_config = {
		/* Total number of DPI queues - set according to the number of cores */
		.nb_queues = 0,
		/* Maximum job size in bytes for RegEx scan match */
		.max_sig_match_len = 500,
		/* Max amount of FIDS per DPI queue */
		.max_packets_per_queue = 100000,
	};

	result = sig_db_init();
	if (result != DOCA_SUCCESS)
		return -1;

	/* Configure RegEx device and queues */
	doca_dpi_config.nb_queues = dpdk_config->port_config.nb_queues;
	dpi_ctx = doca_dpi_init(&doca_dpi_config, &ret);
	if (dpi_ctx == NULL) {
		DOCA_LOG_ERR("DPI init failed, (%d)", result);
		sig_db_destroy();
		return ret;
	}
	ret = doca_dpi_load_signatures(dpi_ctx, ips_config->cdo_filename);
	if (ret < 0) {
		DOCA_LOG_ERR("Loading DPI signature failed");
		doca_dpi_destroy(dpi_ctx);
		sig_db_destroy();
		return ret;
	}

	dpi_worker->dpi_on_match = dpi_worker_on_match;
	dpi_worker->user_data = (void *)ips_config;
	dpi_worker->max_dpi_depth = 0;
	dpi_worker->dpi_ctx = dpi_ctx;
	if (ips_config->netflow_source_id) {
		if (init_netflow_schema_and_source(ips_config->netflow_source_id, "IPS_netflow_metric") != DOCA_SUCCESS) {
			DOCA_LOG_ERR("DOCA Telemetry Netflow init failed");
			doca_dpi_destroy(dpi_ctx);
			sig_db_destroy();
			return -1;
		}
		dpi_worker->send_netflow_record = enqueue_netflow_record_to_ring;
	}
	return 0;
}

void
ips_destroy(struct application_dpdk_config *dpdk_config, struct ips_config *ips)
{
	dpi_worker_lcores_stop(dpi_ctx);

	sig_db_destroy();

	if (ips->netflow_source_id)
		destroy_netflow_schema_and_source();

	doca_dpi_destroy(dpi_ctx);
}

doca_error_t
register_ips_params()
{
	doca_error_t result;
	struct doca_argp_param *print_on_match_param, *netflow_param, *csv_param, *cdo_param, *frag_param;

	/* Create and register print on match param */
	result = doca_argp_param_create(&print_on_match_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create ARGP param: %s", doca_get_error_string(result));
		return result;
	}
	doca_argp_param_set_short_name(print_on_match_param, "p");
	doca_argp_param_set_long_name(print_on_match_param, "print-match");
	doca_argp_param_set_description(print_on_match_param, "Prints FID when matched in DPI engine");
	doca_argp_param_set_callback(print_on_match_param, print_match_callback);
	doca_argp_param_set_type(print_on_match_param, DOCA_ARGP_TYPE_BOOLEAN);
	result = doca_argp_register_param(print_on_match_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to register program param: %s", doca_get_error_string(result));
		return result;
	}

	/* Create and register Netflow param */
	result = doca_argp_param_create(&netflow_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create ARGP param: %s", doca_get_error_string(result));
		return result;
	}
	doca_argp_param_set_short_name(netflow_param, "n");
	doca_argp_param_set_long_name(netflow_param, "netflow");
	doca_argp_param_set_arguments(netflow_param, "<source_id>");
	doca_argp_param_set_description(netflow_param, "Collect netflow statistics and set source_id if value is set");
	doca_argp_param_set_callback(netflow_param, netflow_callback);
	doca_argp_param_set_type(netflow_param, DOCA_ARGP_TYPE_INT);
	result = doca_argp_register_param(netflow_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to register program param: %s", doca_get_error_string(result));
		return result;
	}

	/* Create and register CSV output param */
	result = doca_argp_param_create(&csv_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create ARGP param: %s", doca_get_error_string(result));
		return result;
	}
	doca_argp_param_set_short_name(csv_param, "o");
	doca_argp_param_set_long_name(csv_param, "output-csv");
	doca_argp_param_set_arguments(csv_param, "<path>");
	doca_argp_param_set_description(csv_param, "Path to the output of the CSV file");
	doca_argp_param_set_callback(csv_param, csv_file_callback);
	doca_argp_param_set_type(csv_param, DOCA_ARGP_TYPE_STRING);
	result = doca_argp_register_param(csv_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to register program param: %s", doca_get_error_string(result));
		return result;
	}

	/* Create and register CDO file param */
	result = doca_argp_param_create(&cdo_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create ARGP param: %s", doca_get_error_string(result));
		return result;
	}
	doca_argp_param_set_short_name(cdo_param, "c");
	doca_argp_param_set_long_name(cdo_param, "cdo");
	doca_argp_param_set_arguments(cdo_param, "<path>");
	doca_argp_param_set_description(cdo_param, "Path to CDO file compiled from a valid PDD");
	doca_argp_param_set_callback(cdo_param, cdo_file_callback);
	doca_argp_param_set_type(cdo_param, DOCA_ARGP_TYPE_STRING);
	doca_argp_param_set_mandatory(cdo_param);
	result = doca_argp_register_param(cdo_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to register program param: %s", doca_get_error_string(result));
		return result;
	}

	/* Create and register fragmented param */
	result = doca_argp_param_create(&frag_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create ARGP param: %s", doca_get_error_string(result));
		return result;
	}
	doca_argp_param_set_short_name(frag_param, "f");
	doca_argp_param_set_long_name(frag_param, "fragmented");
	doca_argp_param_set_description(frag_param, "Enables processing fragmented packets");
	doca_argp_param_set_callback(frag_param, frag_callback);
	doca_argp_param_set_type(frag_param, DOCA_ARGP_TYPE_BOOLEAN);
	result = doca_argp_register_param(frag_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to register program param: %s", doca_get_error_string(result));
		return result;
	}

	/* Register version callback for DOCA SDK & RUNTIME */
	result = doca_argp_register_version_callback(sdk_version_callback);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to register version callback: %s", doca_get_error_string(result));
		return result;
	}
	return DOCA_SUCCESS;
}
