/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef L2_REFLECTOR_CORE_H_
#define L2_REFLECTOR_CORE_H_

#include <infiniband/mlx5dv.h>
#include <libflexio/flexio.h>
#include <../common/l2_reflector_common.h>

/* L2 Reflector configuration structure */
struct l2_reflector_config {
	char 				*device_name;		/* IB device name */
	struct l2_reflector_data	*dev_data;		/* device data */

	/* IB Verbs resources. */
	struct ibv_context		*ibv_ctx;		/* IB device context */
	struct ibv_pd			*pd;			/* Protection domain */
	struct mlx5dv_devx_uar		*uar;			/* User Access Region */

	/* FlexIO resources */
	flexio_uintptr_t		dev_data_daddr;		/* Data address accessible by the device */
	struct flexio_process		*flexio_process;	/* FlexIO process */
	struct flexio_outbox		*flexio_outbox;		/* FlexIO outbox */
	struct flexio_uar		*flexio_uar;		/* FlexIO UAR */
	struct flexio_event_handler	*event_handler;		/* Event handler on device */

	struct app_transfer_cq		rq_cq_transf;
	struct app_transfer_cq		sq_cq_transf;

	struct flexio_mkey		*rqd_mkey;
	struct app_transfer_wq		rq_transf;

	struct flexio_mkey		*sqd_mkey;
	struct app_transfer_wq		sq_transf;

	struct flexio_cq 		*flexio_rq_cq_ptr;	/* FlexIO RQ CQ */
	struct flexio_cq 		*flexio_sq_cq_ptr;	/* FlexIO SQ CQ */
	struct flexio_rq 		*flexio_rq_ptr;		/* FlexIO RQ */
	struct flexio_sq 		*flexio_sq_ptr;		/* FlexIO SQ */

	/* mlx5dv direct rules resources, used for steering rules */
	struct mlx5dv_dr_domain		*dr_domain;		/* DR domain object */
	struct mlx5dv_dr_table		*dr_table;		/* DR table in the domain at specific level */
	struct mlx5dv_dr_matcher	*dr_matcher;		/* DR matcher object in the table */
	struct mlx5dv_dr_action		*tir_action;		/* Rule action */
	struct mlx5dv_dr_rule		*dr_rule;		/* Steering rule */
};

/*
 * Open IB device context and allocate UAR and PD
 *
 * @app_cfg [in]: application configuration structure
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
doca_error_t l2_reflector_setup_ibv_device(struct l2_reflector_config *app_cfg);

/*
 * Allocate FlexIO process and device resources
 *
 * @app_cfg [in]: application configuration structure
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
doca_error_t l2_reflector_setup_device(struct l2_reflector_config *app_cfg);

/*
 * Allocate device memory and WQs
 *
 * @app_cfg [in]: application configuration structure
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
doca_error_t l2_reflector_allocate_device_resources(struct l2_reflector_config *app_cfg);

/*
 * Create steering rule that sends all packets to the device
 *
 * @app_cfg [in]: application configuration structure
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
doca_error_t l2_reflector_create_steering_rule(struct l2_reflector_config *app_cfg);

/*
 * Destroy WQs and free device memory
 *
 * @app_cfg [in]: application configuration structure
 */
void l2_reflector_device_resources_destroy(struct l2_reflector_config *app_cfg);

/*
 * Destroy steering rule resources
 *
 * @app_cfg [in]: application configuration structure
 */
void l2_reflector_steering_rule_destroy(struct l2_reflector_config *app_cfg);

/*
 * Destroy FlexIO process and device resources
 *
 * @app_cfg [in]: application configuration structure
 */
void l2_reflector_device_destroy(struct l2_reflector_config *app_cfg);

/*
 * Destroy IB device context
 *
 * @app_cfg [in]: application configuration structure
 */
void l2_reflector_ibv_device_destroy(struct l2_reflector_config *app_cfg);

/*
 * L2 Reflector destroy
 *
 * @app_cfg [in]: application configuration structure
 */
void l2_reflector_destroy(struct l2_reflector_config *app_cfg);

/*
 * Register the command line parameters for the IPS application
 *
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
doca_error_t register_l2_reflector_params(void);

#endif
